from typing import List, Optional, Dict

from django.core.exceptions import FieldError
from django.db.models import ForeignKey, AutoField, BooleanField, BigAutoField, BigIntegerField, FloatField, \
    IntegerField, NullBooleanField, PositiveIntegerField, PositiveSmallIntegerField

from ..searchset.fields.django import DjangoSearchField, DjangoSearchFieldWithoutWildcard
from ..searchset.mapping import DjangoMapping
from ..searchset.utils import FieldType
from ..parser import LuceneToDjangoParserMixin

from .base import BaseSearchSet
from .fields.django import default_django_field_types_to_fields


django_model_field_to_field_type = {
    AutoField: FieldType.INTEGER,
    BigAutoField: FieldType.INTEGER,
    BigIntegerField: FieldType.INTEGER,
    BooleanField: FieldType.BOOLEAN,
    FloatField: FieldType.FLOAT,
    IntegerField: FieldType.INTEGER,
    NullBooleanField: FieldType.NULL_BOOLEAN,
    PositiveIntegerField: FieldType.INTEGER,  # todo positive integer field
    PositiveSmallIntegerField: FieldType.INTEGER,
}


class DjangoSearchSet(LuceneToDjangoParserMixin, BaseSearchSet):
    _field_base_class = DjangoSearchFieldWithoutWildcard
    _default_field = DjangoSearchField

    _field_type_to_field_class = default_django_field_types_to_fields
    _django_model_field_to_field_type = django_model_field_to_field_type

    _field_sources: Optional[List[str]] = None

    @classmethod
    def filter(cls, queryset, search_terms):
        query = cls.parse(raw_expression=search_terms)
        try:
            return queryset.filter(query)
        except FieldError:
            return queryset.none()

    # for search helper

    _mapping_class = DjangoMapping

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        return {field.name: cls._django_model_field_to_field_type.get(field.__class__)
                for field in cls.Meta.model._meta.fields
                if not isinstance(field, ForeignKey)}

    @classmethod
    def get_fields_sources(cls) -> List[str]:
        """
        Returns sources for field defined in searchset class
        """
        if cls._field_sources is None:
            cls._field_sources = list()

            for name, _cls in cls.__dict__.items():
                if isinstance(_cls, cls._field_base_class):
                    cls._field_sources.extend(_cls.get_sources(name))

        return cls._field_sources



from rest_framework_elasticsearch.es_filters import BaseEsFilterBackend

from ..utils import LuceneSearchException


class ElasticLuceneSearchFilterMixin(BaseEsFilterBackend):
    def filter_search(self, request, search, view):
        search_terms = self.get_base_search_terms(request)

        # if there nothing to search - return
        if not search_terms:
            return search

        searchset_class = self.get_searchset_class(view, request)
        filtered_search = self.lucene_filter_search(searchset_class=searchset_class, search_terms=search_terms,
                                                    search=search)

        if filtered_search is None:
            return search.filter('query_string', **{'query': search_terms})

        return filtered_search

    def lucene_filter_search(self, searchset_class, search_terms, search):
        if searchset_class is not None:
            try:
                return searchset_class.filter(search, search_terms)
            except LuceneSearchException:
                return None
        return None



from lucyparser import parse
from lucyparser.exceptions import BaseLucyException
from lucyparser.tree import BaseNode
from lucyparser.parsing import Parser as LucyParser

from ..utils import LuceneSearchException


class CyrillicParser(LucyParser):
    # with redefined `permitted_name_value_char` it's only used for error messages
    value_chars = "letter, digit, or one of -.*_?!;,:@|"

    def permitted_name_value_char(self, c: str):
        return c.isalnum() or c in "-.*_?!;,:@|"


class BaseLuceneParserMixin:
    @classmethod
    def parse(cls, raw_expression: str):
        """
        Parses raw expression to query tree
        """

        try:
            tree = parse(string=raw_expression, parser_class=CyrillicParser)
        except BaseLucyException:
            raise LuceneSearchException()

        parsed_tree = cls._parse_tree(tree=tree)

        if parsed_tree is None:
            raise LuceneSearchException()

        return parsed_tree

    @classmethod
    def _parse_tree(cls, tree: BaseNode):
        """
        Parses lucyparsers tree to query tree
        """
        raise NotImplementedError()



from typing import List

from .base import MappingValue, Mapping


class DjangoMappingValue(MappingValue):
    def _get_values(self, qs, prefix: str) -> List[str]:
        result_qs = qs.filter(**{'{}__icontains'.format(self.sources[0]): prefix}).\
            values_list(self.sources[0], flat=True).distinct()

        other_qss = [
            qs.filter(**{'{}__icontains'.format(source): prefix}).values_list(source, flat=True).distinct()
            for source in self.sources[1:]
        ]

        if other_qss:
            result_qs = result_qs.union(*other_qss)

        return list(set(result_qs))


class DjangoMapping(Mapping):
    _value_class = DjangoMappingValue



class LuceneSearchFilter:
    def get_base_search_terms(self, request) -> str:
        """
        Search terms are set by a ?search=... query parameter,
        and may be comma and/or whitespace delimited.
        """
        return request.query_params.get(self.search_param, '')

    def get_searchset_class(self, view, request):
        """
        Returns searchset class if it presented in view
        """
        return getattr(view, 'search_class', None)



from django.db.models import Q
from lucyparser.tree import ExpressionNode, AndNode, OrNode, NotNode

from .base import BaseLuceneParserMixin


class LuceneToDjangoParserMixin(BaseLuceneParserMixin):
    @classmethod
    def _parse_tree(cls, tree):
        if isinstance(tree, ExpressionNode):
            return cls.get_query_for_field(tree)

        is_and_node = isinstance(tree, AndNode)
        is_or_node = isinstance(tree, OrNode)
        is_not_node = isinstance(tree, NotNode)

        query = Q()

        if is_and_node or is_not_node or is_or_node:
            queries = [cls._parse_tree(tree=child) for child in tree.children]
            queries = [q for q in queries if q is not None]

            if is_and_node:
                for q in queries:
                    query = query & q
                query = (query)

            elif is_or_node:
                for q in queries:
                    query = query | q
                query = (query)

            elif is_not_node:
                for q in queries:
                    query = query & q
                query = ~Q(query)

        return query



from typing import Dict

from ..searchset.fields.elastic import ElasticSearchField
from ..searchset.mapping import ElasticMapping
from ..searchset.utils import FieldType
from ..parser import LuceneToElasticParserMixin

from .base import BaseSearchSet
from .fields.elastic import default_eclipse_field_types_to_fields


elastic_data_type_to_field_type = {
    "long": FieldType.INTEGER,
    "integer": FieldType.INTEGER,
    "short": FieldType.INTEGER,
    "double": FieldType.FLOAT,
    "float": FieldType.FLOAT,
    "boolean": FieldType.NULL_BOOLEAN,
}


class ElasticSearchSet(LuceneToElasticParserMixin, BaseSearchSet):
    _field_base_class = ElasticSearchField
    _default_field = ElasticSearchField

    _field_type_to_field_class = default_eclipse_field_types_to_fields
    _elastic_data_type_to_field_type = elastic_data_type_to_field_type

    @classmethod
    def filter(cls, search, search_terms):
        query = cls.parse(raw_expression=search_terms)
        return None if query is None else search.query(query)

    # for search helper

    _mapping_class = ElasticMapping

    @classmethod
    def _format_mapping_values(cls, mapping, prefix="") -> Dict[str, FieldType]:
        field_name_to_field_type = dict()

        for key, value in mapping.items():
            if "properties" in value:
                field_name_to_field_type.update(cls._format_mapping_values(value["properties"], ".".join([prefix, key]) if prefix else key))
            else:
                field_name = ".".join([prefix, key]) if prefix else key
                field_name_to_field_type[field_name] = cls._elastic_data_type_to_field_type.get(value.get("type"))

        return field_name_to_field_type

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        model_instance = cls.Meta.model()
        index_to_mapping = cls.Meta.model._get_es_client().indices.get_mapping(index=model_instance._get_index())
        if not index_to_mapping:
            return dict()

        last_index = max(index_to_mapping)

        try:
            last_mapping = index_to_mapping[last_index]['mappings'][model_instance._doc_type.name]['properties']
        except KeyError:
            return dict()

        field_name_to_field_type = cls._format_mapping_values(last_mapping)
        return field_name_to_field_type
